#  Copyright 2020-2024 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
import configparser
import logging
import os
import uuid
from typing import Optional
from collections.abc import Sequence

from skytemple.core.async_tasks.delegator import AsyncConfiguration
from skytemple_files.common.impl_cfg import ImplementationType
from skytemple_files.common.project_file_manager import ProjectFileManager
from skytemple_files.common.util import open_utf8

CONFIG_FILE_NAME = "config.ini"

SECT_GENERAL = "General"
SECT_WINDOW = "Window"
SECT_RECENT_FILES = "Recent"
SECT_INTEGRATION_DISCORD = "Discord"

KEY_RECENT_1 = "file1"
KEY_RECENT_2 = "file2"
KEY_RECENT_3 = "file3"
KEY_RECENT_4 = "file4"
KEY_RECENT_5 = "file5"

KEY_ASSISTANT_SHOWN = "assistant_shown"
KEY_GTK_THEME = "gtk_theme"
KEY_LOCALE = "locale"
KEY_USE_NATIVE_FILE_HANDLERS = "use_native_file_handlers"
KEY_ASYNC_CONFIGURATION = "async_configuration"
KEY_ALLOW_SENTRY = "send_error_reports"
KEY_SENTRY_USER_ID = "error_reports_user_id"
KEY_ENABLE_CSD = "enable_csd"
KEY_APPROVED_PLUGINS = "approved_plugins"
KEY_SHOW_SYMBOLS_SCREEN_WARNING = "symbols_warning"

KEY_WINDOW_SIZE_X = "width"
KEY_WINDOW_SIZE_Y = "height"
KEY_WINDOW_POS_X = "pos_x"
KEY_WINDOW_POS_Y = "pos_y"
KEY_WINDOW_IS_MAX = "is_max"

KEY_INTEGRATION_DISCORD_DISCORD_ENABLED = "enabled"
logger = logging.getLogger(__name__)


class SkyTempleSettingsStore:
    def __init__(self):
        self.config_dir = os.path.join(ProjectFileManager.shared_config_dir())
        os.makedirs(self.config_dir, exist_ok=True)
        self.config_file = os.path.join(self.config_dir, CONFIG_FILE_NAME)
        self.loaded_config = configparser.ConfigParser()
        if os.path.exists(self.config_file):
            try:
                with open_utf8(self.config_file, "r") as f:
                    self.loaded_config.read_file(f)
            except BaseException as err:
                logger.error("Error reading config, falling back to default.", exc_info=err)

    def get_recent_files(self) -> list[str]:
        recents = []
        if SECT_RECENT_FILES in self.loaded_config:
            if KEY_RECENT_1 in self.loaded_config[SECT_RECENT_FILES]:
                recents.append(self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_1])
            if KEY_RECENT_2 in self.loaded_config[SECT_RECENT_FILES]:
                recents.append(self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_2])
            if KEY_RECENT_3 in self.loaded_config[SECT_RECENT_FILES]:
                recents.append(self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_3])
            if KEY_RECENT_4 in self.loaded_config[SECT_RECENT_FILES]:
                recents.append(self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_4])
            if KEY_RECENT_5 in self.loaded_config[SECT_RECENT_FILES]:
                recents.append(self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_5])
        return recents

    def set_recent_files(self, recent_files: list[str]):
        # We only save the last 5 opened files. If the list is smaller than the old list, old entries are kept.
        if SECT_RECENT_FILES not in self.loaded_config:
            self.loaded_config[SECT_RECENT_FILES] = {}
        if len(recent_files) > 0:
            self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_1] = recent_files[0]
        if len(recent_files) > 1:
            self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_2] = recent_files[1]
        if len(recent_files) > 2:
            self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_3] = recent_files[2]
        if len(recent_files) > 3:
            self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_4] = recent_files[3]
        if len(recent_files) > 4:
            self.loaded_config[SECT_RECENT_FILES][KEY_RECENT_5] = recent_files[4]
        self._save()

    def get_assistant_shown(self) -> bool:
        if SECT_GENERAL in self.loaded_config:
            if KEY_ASSISTANT_SHOWN in self.loaded_config[SECT_GENERAL]:
                return int(self.loaded_config[SECT_GENERAL][KEY_ASSISTANT_SHOWN]) > 0
        return False

    def set_assistant_shown(self, value: bool):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_ASSISTANT_SHOWN] = "1" if value else "0"
        self._save()

    def get_gtk_theme(self, default=None) -> str:
        if SECT_GENERAL in self.loaded_config:
            if KEY_GTK_THEME in self.loaded_config[SECT_GENERAL]:
                return self.loaded_config[SECT_GENERAL][KEY_GTK_THEME]
        return default

    def set_gtk_theme(self, value: str):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_GTK_THEME] = value
        self._save()

    def get_locale(self, default="") -> str:
        if SECT_GENERAL in self.loaded_config:
            if KEY_LOCALE in self.loaded_config[SECT_GENERAL]:
                return self.loaded_config[SECT_GENERAL][KEY_LOCALE]
        return default

    def set_locale(self, value: str):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_LOCALE] = value
        self._save()

    def get_window_size(self) -> Optional[tuple[int, int]]:
        if SECT_WINDOW in self.loaded_config:
            if (
                KEY_WINDOW_SIZE_X in self.loaded_config[SECT_WINDOW]
                and KEY_WINDOW_SIZE_Y in self.loaded_config[SECT_WINDOW]
            ):
                return int(self.loaded_config[SECT_WINDOW][KEY_WINDOW_SIZE_X]), int(
                    self.loaded_config[SECT_WINDOW][KEY_WINDOW_SIZE_Y]
                )
        return None

    def set_window_size(self, dim: tuple[int, int]):
        if SECT_WINDOW not in self.loaded_config:
            self.loaded_config[SECT_WINDOW] = {}
        self.loaded_config[SECT_WINDOW][KEY_WINDOW_SIZE_X] = str(dim[0])
        self.loaded_config[SECT_WINDOW][KEY_WINDOW_SIZE_Y] = str(dim[1])
        self._save()

    def get_window_position(self) -> Optional[tuple[int, int]]:
        if SECT_WINDOW in self.loaded_config:
            if (
                KEY_WINDOW_POS_X in self.loaded_config[SECT_WINDOW]
                and KEY_WINDOW_POS_Y in self.loaded_config[SECT_WINDOW]
            ):
                return int(self.loaded_config[SECT_WINDOW][KEY_WINDOW_POS_X]), int(
                    self.loaded_config[SECT_WINDOW][KEY_WINDOW_POS_Y]
                )
        return None

    def set_window_position(self, pos: tuple[int, int]):
        if SECT_WINDOW not in self.loaded_config:
            self.loaded_config[SECT_WINDOW] = {}
        self.loaded_config[SECT_WINDOW][KEY_WINDOW_POS_X] = str(pos[0])
        self.loaded_config[SECT_WINDOW][KEY_WINDOW_POS_Y] = str(pos[1])
        self._save()

    def get_window_maximized(self) -> bool:
        if SECT_WINDOW in self.loaded_config:
            if (
                KEY_WINDOW_IS_MAX in self.loaded_config[SECT_WINDOW]
                and self.loaded_config[SECT_WINDOW][KEY_WINDOW_IS_MAX] == "True"
            ):
                return True
        return False

    def set_window_maximized(self, value: bool):
        if SECT_WINDOW not in self.loaded_config:
            self.loaded_config[SECT_WINDOW] = {}
        self.loaded_config[SECT_WINDOW][KEY_WINDOW_IS_MAX] = str(value)
        self._save()

    def get_integration_discord_enabled(self) -> bool:
        if SECT_INTEGRATION_DISCORD in self.loaded_config:
            if KEY_INTEGRATION_DISCORD_DISCORD_ENABLED in self.loaded_config[SECT_INTEGRATION_DISCORD]:
                return int(self.loaded_config[SECT_INTEGRATION_DISCORD][KEY_INTEGRATION_DISCORD_DISCORD_ENABLED]) > 0
        return True  # default is enabled.

    def set_integration_discord_enabled(self, value: bool):
        if SECT_INTEGRATION_DISCORD not in self.loaded_config:
            self.loaded_config[SECT_INTEGRATION_DISCORD] = {}
        self.loaded_config[SECT_INTEGRATION_DISCORD][KEY_INTEGRATION_DISCORD_DISCORD_ENABLED] = "1" if value else "0"
        self._save()

    def get_allow_sentry(self) -> bool:
        if SECT_GENERAL in self.loaded_config:
            if KEY_ALLOW_SENTRY in self.loaded_config[SECT_GENERAL]:
                return int(self.loaded_config[SECT_GENERAL][KEY_ALLOW_SENTRY]) > 0
        return False  # default is disabled.

    def set_allow_sentry(self, value: bool):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_ALLOW_SENTRY] = "1" if value else "0"
        self._save()

    def getset_sentry_user_id(self) -> str:
        if SECT_GENERAL in self.loaded_config:
            if KEY_SENTRY_USER_ID in self.loaded_config[SECT_GENERAL]:
                return self.loaded_config[SECT_GENERAL][KEY_SENTRY_USER_ID]
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_SENTRY_USER_ID] = str(uuid.uuid4())
        self._save()
        return self.loaded_config[SECT_GENERAL][KEY_SENTRY_USER_ID]

    def is_allow_sentry_set(self) -> bool:
        if SECT_GENERAL in self.loaded_config:
            if KEY_ALLOW_SENTRY in self.loaded_config[SECT_GENERAL]:
                return True
        return False

    def get_implementation_type(self) -> ImplementationType:
        if SECT_GENERAL in self.loaded_config:
            if KEY_USE_NATIVE_FILE_HANDLERS in self.loaded_config[SECT_GENERAL]:
                if int(self.loaded_config[SECT_GENERAL][KEY_USE_NATIVE_FILE_HANDLERS]) <= 0:
                    return ImplementationType.PYTHON
        return ImplementationType.NATIVE

    def set_implementation_type(self, value: ImplementationType):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_USE_NATIVE_FILE_HANDLERS] = (
            "1" if value == ImplementationType.NATIVE else "0"
        )
        self._save()

    def get_async_configuration(self) -> AsyncConfiguration:
        # Not supported to be configured anymore.
        return AsyncConfiguration.default()

    def set_async_configuration(self, _value: AsyncConfiguration):
        # Not supported to be configured anymore.
        pass

    def csd_enabled(self) -> bool:
        if SECT_GENERAL in self.loaded_config:
            if KEY_ENABLE_CSD in self.loaded_config[SECT_GENERAL]:
                try:
                    return bool(int(self.loaded_config[SECT_GENERAL][KEY_ENABLE_CSD]))
                except Exception:
                    return True
        return True

    def set_csd_enabled(self, value: bool):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_ENABLE_CSD] = "1" if value else "0"
        self._save()

    def get_approved_plugins(self) -> Sequence[str]:
        if SECT_GENERAL in self.loaded_config:
            if KEY_APPROVED_PLUGINS in self.loaded_config[SECT_GENERAL]:
                return self.loaded_config[SECT_GENERAL][KEY_APPROVED_PLUGINS].split(",")
        return []

    def set_approved_plugins(self, plugin_names: Sequence[str]):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_APPROVED_PLUGINS] = ",".join(plugin_names)
        self._save()

    def get_show_symbols_screen_warning(self) -> bool:
        if SECT_GENERAL in self.loaded_config:
            if KEY_SHOW_SYMBOLS_SCREEN_WARNING in self.loaded_config[SECT_GENERAL]:
                return int(self.loaded_config[SECT_GENERAL][KEY_SHOW_SYMBOLS_SCREEN_WARNING]) > 0
        return True

    def set_show_symbols_screen_warning(self, value: bool):
        if SECT_GENERAL not in self.loaded_config:
            self.loaded_config[SECT_GENERAL] = {}
        self.loaded_config[SECT_GENERAL][KEY_SHOW_SYMBOLS_SCREEN_WARNING] = "1" if value else "0"
        self._save()

    def _save(self):
        with open_utf8(self.config_file, "w") as f:
            self.loaded_config.write(f)
